from eincheck.parser.grammar import parse_shape_spec

parser_cache_info = parse_shape_spec.cache_info
parser_cache_clear = parse_shape_spec.cache_clear
parser_resize_cache = parse_shape_spec.reset_maxsize
